from airflow.providers.slack.operators.slack_webhook import SlackWebhookOperator
from airflow.configuration import conf


SLACK_CONN_ID = "slack_airflow"

DAG_OWNER_MAPPING = {
    "management": "U02TS6TVBP1", #Charles
    "finops": "U06CCADAA21", #Victor
    "billing": "U06CCADAA21", #Victor
    "monitoring": "U08B0Q00T9U", #Cris
    "security": "U02TS6TVBP1", #Charles
    "chatops": "U04PS4C43EC", #Nawfel
    "gitops": "U085HU06BQ9", #Virginia
    "sales_marketing": "U02UKEM1XFB", #Andrea
    "web_analytics": "U06U5P7SV7G", #Greta
    "internal_doc": "U04JMG5CU0K" #Alex
}

def task_fail_slack_alert(context):
    webserver_url = conf.get('webserver', 'base_url')
    dag_id = context.get("task_instance").dag_id
    task_id = context.get("task_instance").task_id
    run_id = context.get("task_instance").run_id
    execution_date = context.get("execution_date")
    dag_url = f"{webserver_url}/dags/{dag_id}/grid?execution_date={execution_date.isoformat()}&tab=graph&dag_run_id={run_id}"

    owner_tag = ""
    for prefix, user_id in DAG_OWNER_MAPPING.items():
        if dag_id.startswith(prefix):
            owner_tag = f"<@{user_id}>"
            break

    slack_msg = """
:red_circle: Task Failed.
*Task*: {task}  
*Dag*: {dag} 
*Execution Time*: {exec_date}
*Dag URL*: <{url}|View in Airflow>
*Owner*: {owner}
""".format(
        task=task_id,
        dag=dag_id,
        exec_date=execution_date,
        url=dag_url,
        owner=owner_tag
    )
    failed_alert = SlackWebhookOperator(
        task_id="slack_fail_alert", 
        slack_webhook_conn_id=SLACK_CONN_ID,
        message=slack_msg,
    )
    return failed_alert.execute(context=context)



from kubernetes.client import models as k8s
from airflow.providers.cncf.kubernetes.operators.pod import KubernetesPodOperator
from airflow.sdk import Variable, Param
from datetime import datetime, timedelta

class ConfigVars:

    def __init__(self):
        COMMON_GITLAB_PAT = Variable.get("COMMON_GITLAB_PAT")
        env_vars = [
            set_env_vars(name="STG_PROJECT_ID", key="STG_PROJECT_ID"),
            set_env_vars(name="LZ_PROJECT_ID", key="LZ_PROJECT_ID"),
            set_env_vars(name="DM_PROJECT_ID", key="DM_PROJECT_ID"),
            set_env_vars(name="DW_PROJECT_ID", key="DW_PROJECT_ID"),
            set_env_vars(name="COMMON_GITLAB_PAT", value=COMMON_GITLAB_PAT)
        ]
        self.env_vars = env_vars


class DbtRunDefaults:
    image_pull_policy = "Always"
    cmds = ["/bin/bash", "-c"]

    get_logs = True
    service_account_name = 'dbt-ksa'
    namespace = "dbt"
    affinity = {'nodeAffinity': {'requiredDuringSchedulingIgnoredDuringExecution':
        {'nodeSelectorTerms': [
            {'matchExpressions': [
                {'key': 'node_pool',
                 'operator': 'In',
                 'values': ['dbt']
                 }
            ]
            }
        ]
        }
    }
    }

    def __init__(self, data_product, environment):
        self.dbt_target_gke = f'{data_product}-{environment}-ci-airflow'
        self.dbt_target = f""" --target="{self.dbt_target_gke}" """
        self.run_id = '{{ run_id }}'
        self.dag_ts = '{{ ts }}'
        self.dbt_profile = f""" --profiles-dir=/app/artifacts"""
        self.dbt_vars = f""" --vars '{{"dag_id": "{self.run_id}" , "dag_ts": "{self.dag_ts}", "orchestrator": "Airflow", 
        "job_run_id": "{self.run_id}"}}'"""
        self.dbt_args = f"""{self.dbt_vars}{self.dbt_target}{self.dbt_profile}"""
        self.docker_image = "europe-west1-docker.pkg.dev/prj-astrafy-artifacts/mds"

def default_dag(data_product, ENV, **kwargs):
    dag = {
        "dag_id": f'{data_product}-k8s-dag-{ENV}',
        "schedule": '0 10 * * *',
        "description": f'Dag for the {data_product} data product - {ENV}',
        "max_active_tasks": 10,
        "catchup": False,
        "is_paused_upon_creation": True,
        "tags": ['k8s', data_product],
        "max_active_runs": 1,
        "dagrun_timeout": timedelta(seconds=36000),
        "sla_miss_callback": None,
        "params": {"is_full_refresh": Param(False, type="boolean")},
        "start_date": datetime(2022, 10, 10, 1, 0),
        "doc_md": __doc__,
    }
    
    dag.update(kwargs)
    
    return dag


def cd_to_local_package(package: str) -> str:
    return f"cd /app/data_products/{package}/"


def set_env_vars(name: str, value=None, key=None):
    if value:
        env_var = k8s.V1EnvVar(name=name, value=value)
        return env_var
    if key:
        env_var = k8s.V1EnvVar(name=name, value_from=k8s.V1EnvVarSource(
            config_map_key_ref=k8s.V1ConfigMapKeySelector(name="internal-data-dbt-config-map", key=key)))
        return env_var

def setup_env_vars(data_product, environment, additional_vars=None):
    """
    Set up environment variables for a given data product and environment,
    with the option to add additional variables.
    """
    config_vars = ConfigVars()

    # Extend with additional environment variables specific to the DAG
    additional_env_vars = [
        set_env_vars(name="DATA_PRODUCT", value=data_product),
        set_env_vars(name="ENV", value=environment)
    ]

    if additional_vars:
        additional_env_vars.extend([
            set_env_vars(name=var_name, value=var_value) for var_name, var_value in additional_vars.items()
        ])

    config_vars.env_vars.extend(additional_env_vars)
    return config_vars.env_vars

def construct_dbt_run_command(dbt_defaults, selector, mode):
    """
    Construct a DBT command using the DbtRunDefaults settings, a selector, and a specified mode.
    """
    dbt_selector = f"--selector {selector}"
    return f"dbt run {dbt_defaults.dbt_args} {dbt_selector} {mode};exit $?"

def create_dbt_task(task_id, name, command, dbt_defaults, env_vars, data_product):
    """
    Create a KubernetesPodOperator task for running DBT commands.
    """
    DP_VERSION = Variable.get(f"DP_{data_product.upper()}_VERSION")
    return KubernetesPodOperator(
        task_id=task_id,
        name=name,
        image=f"{dbt_defaults.docker_image}/dbt/{data_product}:{DP_VERSION}",
        cmds=dbt_defaults.cmds,
        arguments=[command],
        image_pull_policy=dbt_defaults.image_pull_policy,
        namespace=dbt_defaults.namespace,
        get_logs=dbt_defaults.get_logs,
        service_account_name=dbt_defaults.service_account_name,
        affinity=dbt_defaults.affinity,
        env_vars=env_vars,
    )


#TO BE REMOVED
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))
from airflow.providers.airbyte.operators.airbyte import AirbyteTriggerSyncOperator
from airflow.sdk import DAG
from astrafy_environment import AstrafyEnvironment


def airbyte_sync(task_id,env: AstrafyEnvironment,dag: DAG = None,) ->AirbyteTriggerSyncOperator:
    dag = dag
    async_source_destination = AirbyteTriggerSyncOperator(
        task_id=task_id,
        airbyte_conn_id=env.airflow_connection_name,
        connection_id=env.airbyte_conn_id,
        asynchronous=False,
        timeout=3600,
        wait_seconds=3,
        dag=dag
        )
    return async_source_destination


from astrafy_environment import AstrafyEnvironment

def get_gcs_bucket_path(environment: AstrafyEnvironment) -> str:
    """
    Constructs the GCS bucket path using the environment configuration.
    
    @param environment: AstrafyEnvironment instance with the configuration
    @return: The constructed GCS bucket path
    """
    return f"gs://{environment.bucket_folder_results}"

def upload_to_gcs(environment: AstrafyEnvironment, target_path: str = "") -> str:
    """
    Generates command to upload dbt artifacts to GCS.
    
    @param environment: AstrafyEnvironment instance with the configuration
    @param target_path: Optional target path within the /app directory
    @return: The bash command string
    """
    bucket_path = get_gcs_bucket_path(environment)
    
    # If target_path is provided, use it in the path, otherwise use the default.
    path = f"/app/{target_path}/target" if target_path else "/app/target"
    
    return f"""
    gcloud storage cp {path}/*.json {bucket_path}/
    """

def download_failed_models(environment: AstrafyEnvironment, target_path: str = "") -> str:
    """
    Generates command to download previous dbt artifacts from GCS.
    
    @param environment: AstrafyEnvironment instance with the configuration
    @return: The bash command string
    """
    bucket_path = get_gcs_bucket_path(environment)
    path = f"/app/{target_path}/target" if target_path else "/app/target"
    return f"""
    mkdir -p {path}
    gcloud storage cp -r {bucket_path}/* /app/target/
    """